(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.IdleHands = factory());
})(this, (function () { 'use strict';

  class ConfigManager {

    constructor(configValues) {
      if (!configValues['logoutUrl']) {
        throw new TypeError('logoutUrl must be defined');
      }

      this.applicationId = window.location.hostname;
      this.cancelButtonText = 'Stay Logged In';
      this.containerSelector = 'body';
      this.containerTitle = 'Session Expiration Warning Prompt';
      this.debug = false;
      this.dialogText = 'Your session will expire in %time seconds due to inactivity.';
      this.dialogTextAllowHtml = false;
      this.documentTitle = 'Session Expiration Warning';
      this.duration = 30 * 1000; // 30 seconds;
      this.events =  ['click', 'keypress', 'scroll', 'wheel', 'mousewheel'];
      this.headerText = 'Session Expiration Warning';
      this.heartbeatInterval = 60 * 1000; // 1 minute;
      this.heartbeatUrl = window.location.href;
      this.logoutButtonText = 'Log Out Now';
      this.logoutDocumentTitle = 'Logging out...';
      this.logoutText = 'Logging out...';
      this.maximumIdleDuration = 60 * 1000 * 60; // 1 hour;
      this.shiftFocus = true;
      this.zIndex = 999;

      for (const key in configValues) {
        this[key] = configValues[key];
      }

      // this.manualLogoutUrl = this.logoutUrl;
    }

  }

  class Logger {

    constructor(applicationId) {
      if (!applicationId) throw new TypeError('applicationId must be defined');

      this.applicationId = applicationId;
    }

    log(message) {
      console.log(`[IDLE HANDS - ${this.applicationId}] ${message}`);
    }

  }

  class Storage {

    prefix;

    constructor(applicationId) {
      if (!applicationId) throw new TypeError('applicationId must be defined');

      this.prefix = applicationId + '_';
    }

    set(key, value) {
      localStorage.setItem(this.prefix + key, value);
    }

    get(key) {
      return localStorage.getItem(this.prefix + key);
    }

    delete(key) {
      localStorage.removeItem(this.prefix + key);
    }

  }

  class Timer {

    constructor(storage, tick = () => {}) {
      if (!storage) throw new TypeError('storage must be defined');

      this.storage = storage;

      this.setStartTime();

      // this.tickInterval = setInterval(tick, 1000);
      // Start the main timer using a Worker.
      this.tickInterval = new WorkerTimer(tick, 1000);
    }

    getCurrentTime() {
      return (new Date).getTime();
    }

    setStartTime() {
      this.storage.set('startTime', this.getCurrentTime());
    }

    getStartTime() {
      return this.storage.get('startTime');
    }

    clearStartTime() {
      this.storage.delete('startTime');
    }

    getIdleTime() {
      const REAL_IDLE_TIME = this.getCurrentTime() - this.getStartTime();

      return REAL_IDLE_TIME - (REAL_IDLE_TIME % 1000);
    }

    atInterval(interval) {
      return (this.getIdleTime() % interval === 0);
    }

    getTimeRemaining(timeAllowed) {
      return (timeAllowed - this.getIdleTime());
    }

    stop() {
      // clearInterval(this.tickInterval);
      // Stop the worker timer.
      this.tickInterval.stop();
    }

  }

  class Heartbeat {

    constructor(url) {
      this.url = url;
    }

    beat() {
      fetch(this.url);
    }

  }

  class Element {

    element;

    create(
      tagName,
      classNames = [],
      styles = {},
      text = '',
      children = [],
      allowHtml = false,
      id = '',
    ) {
      const ELEMENT = document.createElement(tagName);

      classNames.forEach(function(className) {
        ELEMENT.classList.add(className);
      });

      if (id !== '') {
        ELEMENT.id = id;
      }

      for (const style in styles) {
        ELEMENT.style.setProperty(style, styles[style], 'important');
      }

      if (allowHtml) {
        ELEMENT.innerHTML = text;
      } else {
        ELEMENT.innerText = text;
      }

      children.forEach(function(child) {
        ELEMENT.appendChild(child);
      });

      return ELEMENT;
    }

  }

  class Button extends Element {

    constructor(text) {
      super();

      this.element = this.create(
        'button',
        [],
        {
          'padding': '4px 1.5em',
          'cursor': 'pointer',
          'text-align': 'center',
          'text-decoration': 'none',
          'color': '#333',
          'border': '1px solid #a6a6a6',
          'border-radius': '20em',
          'background-color': '#f2f1eb',
          'background-image': 'linear-gradient(to bottom, #f6f6f3, #e7e7df)',
          'text-shadow': '0 1px hsla(0, 0%, 100%, 0.6)',
          'font-size': '0.875rem',
          'font-weight': '600',
          'line-height': '1.25em'
        },
        text
      );
    }

  }

  class Span extends Element {

    constructor(text = '', children = [], styles = {}, allowHtml = false) {
      super();

      this.element = this.create(
        'span',
        [],
        styles,
        text,
        children,
        allowHtml
      );
    }

  }

  class Header extends Span {

    constructor(text) {
      super(
        text,
        [],
        {
          'display': 'block',
          'font-family': '"Lucida Grande", "Lucida Sans Unicode", "DejaVu Sans", "Lucida Sans", sans-serif',
          'font-size': '20px',
          'font-weight': 'bold',
          'color': '#fff',
          'text-align': 'center',
          'background-color': '#1d4289',
          'padding': '8px',
        }
      );
    }

  }

  class Div extends Element {

    constructor(children, styles = {}) {
      super();

      this.element = this.create(
        'div',
        [],
        styles,
        '',
        children
      );
    }

  }

  class Dialog extends Div {

    header;
    textContainer;
    buttonContainer;

    constructor(
      header,
      textContainer,
      buttonContainer
    ) {
      const CHILDREN = [
        header.element,
        textContainer.element,
        buttonContainer.element,
      ];

      const STYLES = {
        'position': 'fixed',
        'top': '50%',
        'left': '50%',
        'transform': 'translate(-50%, -50%)',
        'max-width': '450px',
        'padding-bottom': '8px',
        'background-color': 'white',
        'color': '#152b22',
        'border': '1px solid #a6a6a6',
      };

      super(CHILDREN, STYLES);

      this.header = header;
      this.textContainer = textContainer;
      this.buttonContainer = buttonContainer;
    }

  }

  class Prompt extends Div {

    dialog;

    constructor(dialog, zIndex) {
      super(
        [dialog.element],
        {
          'position': 'fixed',
          'height': '100%',
          'width': '100%',
          'top': '0',
          'left': '0',
          'background-image': 'url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNksAcAAEUAQRtOwGEAAAAASUVORK5CYII=")',
          'z-index': zIndex,
        }
      );

      this.dialog = dialog;
    }

  }

  class ButtonContainer extends Div {

    cancelButton;
    logoutButton;

    constructor(cancelButton, logoutButton) {
      super(
        [cancelButton.element, logoutButton.element],
        {
          'display': 'flex',
          'justify-content': 'space-evenly',
          'margin-bottom': '10px',
        }
      );

      this.cancelButton = cancelButton;
      this.logoutButton = logoutButton;
    }

  }

  class TextContainer extends Div {

    dialogText;

    constructor(dialogText) {
      super(
        [dialogText.element],
        {
          'margin': '15px 20px',
          'line-height': '19px',
          'font-size': '14px',
          'font-family': '"Lucida Grande", "Lucida Sans Unicode", "DejaVu Sans", "Lucida Sans", sans-serif',
        }
      );

      this.dialogText = dialogText;
    }

  }

  class PromptFactory {

    static create(
      headerText,
      dialogText,
      cancelButtonText,
      logoutButtonText,
      zIndex
    ) {

      const CANCEL_BUTTON = new Button(cancelButtonText);
      const LOGOUT_BUTTON = new Button(logoutButtonText);

      const DIALOG = new Dialog(
        new Header(headerText),
        new TextContainer(dialogText),
        new ButtonContainer(CANCEL_BUTTON, LOGOUT_BUTTON)
    );

      return new Prompt(DIALOG, zIndex);
    }

  }

  class Object extends Element {

    constructor() {
      super();

      this.element = this.create(
        'div',
        ['logout-timeout-container'],
        {
          'display': 'none',
          'position': 'fixed',
          'top': '0',
          'left': '0',
          'height': '100vh',
          'width': '100vw',
          'border': 'none',
          'z-index': '9999',
        },
        '',
        [],
        false,
        'logout-timeout-window',
      );
    }

  }

  class Container extends Object {

    prompt;
    isDisplayed;

    constructor(title, documentTitle, prompt) {
      super();

      this.element.title = title;
      this.prompt = prompt;
      this.isDisplayed = false;

    }

    display() {
      this.element.style.display = 'block';
      this.isDisplayed = 'true';
    }

    hide() {
      this.element.style.display = 'none';
      this.isDisplayed = false;
    }

  }

  class DialogText extends Span {

    time;
    renderedText;

    constructor(dialogText, time, allowHtml) {
      const TEMPLATE_PARTS = dialogText.split('%time');

      if (TEMPLATE_PARTS.length > 2) {
        throw new SyntaxError(
          'dialogText must contain no more than one "%time" placeholder'
        );
      }

      let renderedText;

      if (TEMPLATE_PARTS.length === 2) {
        renderedText = TEMPLATE_PARTS[0] + time + TEMPLATE_PARTS[1];
      } else {
        renderedText = dialogText;
      }

      super(
        renderedText,
        [],
        {},
        allowHtml
      );

      this.time = time;
      this.renderedText = renderedText;
    }

  }

  class IdleHands {

    config;
    originalDocumentTitle;
    logger;
    storage;
    timer;
    heartbeat;
    resetHandler;
    logoutHandler;
    container;
    parent;
    originalActiveElement;

    flashingTitle;
    eventListenersSet;

    constructor(config = {}) {
      this.config = new ConfigManager(config);

      const APPLICATION_ID = this.config.applicationId;

      this.logger = new Logger(APPLICATION_ID);
      this.storage = new Storage(APPLICATION_ID);
      this.timer = this.createTimer();
      this.heartbeat = new Heartbeat(this.config.heartbeatUrl);
      this.resetHandler = this.reset.bind(this);
      this.logoutHandler = this.logOut.bind(this);
      this.originalDocumentTitle = document.title;
      this.originalActiveElement = (document.activeElement || document.body);
      this.parent = document.querySelector(this.config.containerSelector);

      this.documentTitle = this.config.documentTitle;
      this.flashingTitle = 0;

      this.eventListenersSet = false;

      this.storage.set('logoutUrl', this.config.logoutUrl);

      this.setContainerElement();
      this.setDocumentEventListeners();
      this.setCancelButtonEventListener();
      this.setLogoutButtonEventListener();
    }

    setContainerElement() {
      this.container = new Container(
        this.config.containerTitle,
        this.config.documentTitle,
        this.createPromptElement()
      );

      this.parent.appendChild(this.container.element);

      this.container
        .element
        .appendChild(this.container.prompt.element);

      this.container
        .element
        .style
        .fontSize = window.getComputedStyle(document.body, null)
          .getPropertyValue('font-size');
    }

    createPromptElement() {
      return PromptFactory.create(
        this.config.headerText,
        this.createDialogText(),
        this.config.cancelButtonText,
        this.config.logoutButtonText,
        this.config.zIndex
      );
    }

    createDialogText() {
      return new DialogText(
        this.config.dialogText,
        this.timer.getTimeRemaining(this.config.maximumIdleDuration) / 1000,
        this.config.dialogTextAllowHtml
      );
    }

    displayLogoutMessage() {
      this.container
        .prompt
        .dialog
        .header
        .element
        .innerText = this.config.logoutText;

      this.container
        .prompt
        .dialog
        .buttonContainer
        .cancelButton
        .element
        .disabled = true;

      this.container
        .prompt
        .dialog
        .buttonContainer
        .logoutButton
        .element
        .disabled = true;
    }

    hideContainer() {
      this.log('Hiding container...');
      this.container.hide();

      this.log('Removing focus from container...');
      this.container.element.blur();

      this.log('Returning focus to original active element...');
      this.originalActiveElement.focus();
    }

    setDocumentEventListeners() {
      this.config.events.forEach(function(event) {
        this.log(`Setting ${event} event listener on document`);
        document.addEventListener(event, this.resetHandler);
      }.bind(this));
      this.eventListenersSet = true;
    }

    unsetDocumentEventListeners() {
      this.config.events.forEach(function(event) {
        this.log(`Removing ${event} event listener from document`);
        document.removeEventListener(event, this.resetHandler);
      }.bind(this));
      this.eventListenersSet = false;
    }

    setCancelButtonEventListener() {
      this.container
        .prompt
        .dialog
        .buttonContainer
        .cancelButton
        .element
        .addEventListener('click', this.resetHandler);
    }

    setLogoutButtonEventListener() {
      this.container
        .prompt
        .dialog
        .buttonContainer
        .logoutButton
        .element
        .addEventListener('click', function() {
          this.storage.set('logoutUrl', this.config.manualLogoutUrl);
          this.logoutHandler();
        }.bind(this));
    }

    createTimer() {
      return new Timer(this.storage, this.tick.bind(this));
    }

    log(message) {
      if (this.config.debug) this.logger.log(message);
    }

    reset() {
      this.log('Reverting document title...');
      document.title = this.originalDocumentTitle;

      this.setDocumentEventListeners();

      this.hideContainer();

      this.log('Stopping timer...');
      this.timer.stop();

      this.log('Assigning new timer...');
      this.timer = this.createTimer();
    }

    logOut() {
      this.log('Updating document title...');
      document.title = this.config.logoutDocumentTitle;

      this.log('Stopping timer...');
      this.timer.stop();

      this.log('Clearing start time...');
      this.timer.clearStartTime();

      this.log('Displaying logout message...');
      this.displayLogoutMessage();

      this.log('Redirecting to logout URL...');
      window.location.replace(this.storage.get('logoutUrl'));
    }

    tick() {
      const TIMER = this.timer;
      const HEARTBEAT_INTERVAL = this.config['heartbeatInterval'];
      const MAXIMUM_IDLE_TIME = this.config.maximumIdleDuration;
      const TIME_REMAINING = TIMER.getTimeRemaining(MAXIMUM_IDLE_TIME);
      const PROMPT_DURATION = this.config.duration;
      const PROMPT_IS_DISPLAYED = this.container.isDisplayed;
      const DIALOG_TEXT_HTML = this.createDialogText(TIME_REMAINING / 1000)
        .element
        .innerHTML;

      this.log('Tick...');

      if (TIME_REMAINING < 0) {
        this.log('Logging out...');
        this.logoutHandler();

        return;
      }

      this.log('Updating prompt...');

      this.container
        .prompt
        .dialog
        .textContainer
        .dialogText
        .element
        .innerHTML = DIALOG_TEXT_HTML;

      if (TIME_REMAINING > PROMPT_DURATION && PROMPT_IS_DISPLAYED) {
        this.hideContainer();
      }

      if (!PROMPT_IS_DISPLAYED) {
        if (this.flashingTitle > 0) {
          this.log('Clearing flashing title...');
          clearInterval(this.flashingTitle);
          this.log('Resetting document title...');
          document.title = this.originalDocumentTitle;
          this.flashingTitle = 0;
        }

        if (!this.eventListenersSet) {
          this.log('Resetting event listeners...');
          this.setDocumentEventListeners();
        }
      }

      if (TIME_REMAINING <= PROMPT_DURATION && !PROMPT_IS_DISPLAYED) {
        this.log('Unsetting event listeners...');
        this.unsetDocumentEventListeners();

        this.log('Set flashing title...');
        this.flashingTitle = setInterval(flashTitle, 2000, this.originalDocumentTitle, this.documentTitle);

        this.log('Displaying prompt...');
        this.container.display();

        if (this.config.shiftFocus) {
          this.log('Shifting focus to cancel button...');
          this.container
            .prompt
            .dialog
            .buttonContainer
            .cancelButton
            .element
            .focus();
        }
      }

      if (TIME_REMAINING <= 0) {
        this.log('Logging out...');
        this.logOut();
      } else if (TIMER.atInterval(HEARTBEAT_INTERVAL)) {
        this.log('Heartbeat...');
        this.heartbeat.beat();
      }
    }

  }

  function flashTitle(pageTitle, newTitle) {
    document.title == pageTitle
        ? (document.title = newTitle)
        : (document.title = pageTitle);
  }

  class WorkerTimer {
    worker = null;
    constructor(callback, interval) {
      const blob = new Blob([`setInterval(() => postMessage(0), ${interval});`]);
      const workerScript = URL.createObjectURL(blob);
      this.worker = new Worker(workerScript);
      this.worker.onmessage = callback;
    }

    stop() {
      this.worker.terminate();
    }
  }

  return IdleHands;

}));
