<?php

declare(strict_types=1);

/**
 * @file
 * Contains \Drupal\logout_timeout\Form\LogoutTimeoutSettingsForm
 **/

namespace Drupal\logout_timeout\Form;

use Drupal\logout_timeout\Service\LogoutTimeoutService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Form for use with configuration of Logout on Timeout.
 */
class LogoutTimeoutSettingsForm extends ConfigFormBase
{

    /**
     * The logout on timeout service.
     *
     * @var \Drupal\logout_timeout\Service\LogoutTimeoutService
     */
    protected $ltService;

    /**
     * Defines the timeout limit.
     *
     * @var int
     *
     * Maximum 32-bit integer that can be handled by setInterval.
     * See https://developer.mozilla.org/en-US/docs/Web/API/setInterval
     */
    private $maxAllowedTimeout = 2147483;

    /**
     * LogoutTimeoutController constructor.
     *
     * @param LogoutTimeoutService $ltService
     */
    public function __construct(LogoutTimeoutService $ltService)
    {
        $this->logoutTimeoutService = $ltService;
    }

    /**
     * {@inheritdoc}
     */
    public static function create(ContainerInterface $container)
    {
        return new static(
            $container->get('logout_timeout.service'),
        );
    }

    /**
     * {@inheritdoc}
     */
    public function getFormId() : string
    {
        return 'logout_timeout_settings';
    }

    /**
     * {@inheritdoc}
     */
    protected function getEditableConfigNames() : array
    {
        return [
            'logout_timeout.settings'
        ];
    }

    /**
     * {@inheritdoc}
     */
    public function buildForm(array $form, FormStateInterface $form_state) : array
    {

        // Turn off browser based HTML5 form validation in favor of Inline Form Errors core module
        $form['#attributes']['novalidate'] = 'novalidate';

        $config = $this->config('logout_timeout.settings');

        $form['enabled'] = [
            '#type'          => 'checkbox',
            '#title'         => $this->t('Enable Timeout on Logout'),
            '#default_value' => $config->get('enabled'),
            '#weight'        => -20,
            '#description'   => $this->t(
                'Enable the Timeout on Logout module on the site to track user inactivity.'
            ),
        ];

        $form['maxInactivitySeconds'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Timeout for inactivity in seconds'),
            '#default_value' => $config->get('maxInactivitySeconds'),
            '#weight'        => -16,
            '#size'          => 10,
            '#required'      => TRUE,
            '#description'   => $this->t(
                'The length of the inactivity, in seconds. Value is total time before logout
                and includes the time displaying the dialog.'
            ),
        ];

        $form['inactivityDialogDuration'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Time displaying the dialog in seconds'),
            '#default_value' => $config->get('inactivityDialogDuration'),
            '#weight'        => -12,
            '#size'          => 10,
            '#required'      => TRUE,
            '#description'   => $this->t(
                'The number of seconds giving the user the chance to extend an inactive session, with the dialog showing.
                Must be less then the Timeout for inactivity value.'
            ),
        ];

        $form['targetLogoutUrl'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Inactivity redirect URL'),
            '#default_value' => $config->get('targetLogoutUrl'),
            '#weight'        => -8,
            '#required'      => TRUE,
            '#description'   => $this->t(
                'The internal URL to send users when they have logged out.'
            ),
        ];

        $checkSamlAuthDescription = $this->t(
            'Check if a user is logged in using the simpleSAMLphp Authentication module,
            and if so use the simpleSAMLphp Authentication logout process to log them out.'
        );
        if (!$this->logoutTimeoutService->getSimpleSamlPhpAuthStatus()) {
            $checkSamlAuthDescription = $checkSamlAuthDescription . $this->t(
                '<br><strong>You do not have the simpleSAMLphp Authentication module enabled.</strong>
                You must have this module enabled to use this feature.'
            );
        }
        $form['checkSamlAuth'] = [
            '#type'          => 'checkbox',
            '#title'         => $this->t('Use simpleSAMLphp Authentication'),
            '#default_value' => $config->get('checkSamlAuth'),
            '#weight'        => -4,
            '#description'   => $checkSamlAuthDescription,
        ];

        $checkHookEventsDispatchDescription = $this->t(
            'Use the module Hook Events Dispatcher\'s User module to dispatch a logout event when using simpleSAMLphp Authentication.'
        );
        if (!$this->logoutTimeoutService->getHookUserDispatchStatus() && $this->logoutTimeoutService->getSimpleSamlPhpAuthStatus()) {
            $checkHookEventsDispatchDescription = $checkHookEventsDispatchDescription . $this->t(
                '<br><strong>You do not have the User Event Dispatcher module enabled.</strong>
                You must have this module enabled to use this feature.'
            );
        } elseif (!$this->logoutTimeoutService->getHookUserDispatchStatus() && !$this->logoutTimeoutService->getSimpleSamlPhpAuthStatus()) {
            $checkHookEventsDispatchDescription = $checkHookEventsDispatchDescription . $this->t(
                '<br><strong>You do not have the User Event Dispatcher nor the Simple SAML PHP Authentication module enabled.</strong>
                You must have both modules enabled to use this feature.'
            );
        }
        $form['checkHookEventsDispatch'] = [
            '#type' => 'checkbox',
            '#title' => $this->t('Use Hook Event Dispatcher'),
            '#default_value' => $config->get('checkHookEventsDispatch'),
            '#weight' => 0,
            '#description' => $checkHookEventsDispatchDescription,
        ];

        $form['dialogTitle'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Dialog title'),
            '#default_value' => $config->get('dialogTitle'),
            '#weight'        => 4,
            '#description'   => $this->t(
                'The title text for the logout dialog'
            ),
        ];

        $form['dialogMessage'] = [
            '#type'          => 'textarea',
            '#title'         => $this->t('Dialog message'),
            '#default_value' => $config->get('dialogMessage'),
            '#weight'        => 8,
            '#description'   => $this->t(
                'The message displayed in the logout dialog.'
            ),
        ];

        $form['stayLoggedInButtonText'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Continue session button text'),
            '#default_value' => $config->get('stayLoggedInButtonText'),
            '#weight'        => 12,
            '#description'   => $this->t(
                'The text used for the button in the dialog for a user to continue their session.'
            ),
        ];

        $form['logoutNowButtonText'] = [
            '#type'          => 'textfield',
            '#title'         => $this->t('Log out now button text'),
            '#default_value' => $config->get('logoutNowButtonText'),
            '#weight'        => 16,
            '#description'   => $this->t(
                'The text used for the button in the dialog for a user end their session and log out immediately.'
            ),
        ];

        $form['debugJavascript'] = [
            '#type'          => 'checkbox',
            '#title'         => $this->t('Enable Debugging'),
            '#default_value' => $config->get('debugJavascript'),
            '#weight'        => 20,
            '#description'   => $this->t(
                'Enable debug log info to be displayed in the browser console.'
            ),
        ];

        return parent::buildForm($form, $form_state);

    }

    /**
     * {@inheritdoc}
     */
    public function validateForm(array &$form, FormStateInterface $form_state) : void
    {

        $values = $form_state->getValues();

        // maxInactivitySeconds
        $maxInactivitySeconds = $values['maxInactivitySeconds'];
        if ($maxInactivitySeconds <= 60) {
            $form_state->setErrorByName(
                'maxInactivitySeconds',
                $this->t('Timeout for inactivity must be greater than 60 seconds.')
            );
        } else if (
            !is_numeric($maxInactivitySeconds)
            || (int) $maxInactivitySeconds != $maxInactivitySeconds
            || $maxInactivitySeconds <= 60
        ) {
            $form_state->setErrorByName(
                'maxInactivitySeconds',
                $this->t('The timeout must be an integer greater than 60.')
            );
        } else if ($maxInactivitySeconds >= $this->maxAllowedTimeout) {
            $form_state->setErrorByName(
                'maxInactivitySeconds',
                $this->t(
                    'Timeout for inactivity must be an integer less than the maximum timeout of %limit seconds.',
                    ['%limit' => $this->maxAllowedTimeout]
                )
            );
        }

        // inactivityDialogDuration
        $inactivityDialogDuration = $values['inactivityDialogDuration'];
        if (
            !is_numeric($inactivityDialogDuration)
            || (int) $inactivityDialogDuration != $inactivityDialogDuration
            || $inactivityDialogDuration < 60
        ) {
            $form_state->setErrorByName(
                'inactivityDialogDuration',
                $this->t('The modal display time must be an integer greater than or equal to 60.')
            );
        } else if ($inactivityDialogDuration >= $maxInactivitySeconds) {
            $form_state->setErrorByName(
                'inactivityDialogDuration',
                $this->t(
                    'The modal display time must be less than the current timeout for inactivity, %timeout seconds.',
                    ['%timeout' => $maxInactivitySeconds]
                )
            );
        } else if ($inactivityDialogDuration >= $this->maxAllowedTimeout) {
            $form_state->setErrorByName(
                'inactivityDialogDuration',
                $this->t(
                    'The modal display time must be an integer less than the maximum timeout of %limit seconds.',
                    ['%limit' => $this->maxAllowedTimeout]
                )
            );
        }

        // targetLogoutUrl
        $targetLogoutUrl = $values['targetLogoutUrl'];
        if (strpos($targetLogoutUrl, '/') == 0) {
            $form_state->setErrorByName(
                'targetLogoutUrl', 
                $this->t("Redirect URL at logout :targetLogoutUrl must not begin with a '/'", [':targetLogoutUrl' => $targetLogoutUrl])
            );
        }

        // dialogMessage
        $dialogMessage = trim($values['dialogMessage']);
        if (empty($dialogMessage)) {
            $form_state->setValue('dialogMessage', 'Your session is about to expire due to inactivity.');
        }

        // dialogTitle
        $dialogTitle = trim($values['dialogTitle']);
        if (empty($dialogTitle)) {
            $form_state->setValue('dialogTitle', 'Session Expiration Warning');
        }

        // logoutNowButtonText
        $logoutNowButtonText = trim($values['logoutNowButtonText']);
        if (empty($logoutNowButtonText)) {
            $form_state->setValue('logoutNowButtonText', 'Logout Now');
        }

        // stayLoggedInButtonText
        $stayLoggedInButtonText = trim($values['stayLoggedInButtonText']);
        if (empty($stayLoggedInButtonText)) {
            $form_state->setValue('stayLoggedInButtonText', 'Stay Logged In');
        }

        parent::validateForm($form, $form_state);

    }

    /**
     * {@inheritdoc}
     */
    public function submitForm(array &$form, FormStateInterface $form_state) : void
    {

        $values                = $form_state->getValues();
        $logoutTimeoutSettings = $this->config('logout_timeout.settings');

        $logoutTimeoutSettings
            ->set('enabled', $values['enabled'])
            ->set('maxInactivitySeconds', $values['maxInactivitySeconds'])
            ->set('inactivityDialogDuration', $values['inactivityDialogDuration'])
            ->set('targetLogoutUrl', $values['targetLogoutUrl'])
            ->set('checkSamlAuth', $values['checkSamlAuth'])
            ->set('checkHookEventsDispatch', $values['checkHookEventsDispatch'])
            ->set('dialogTitle', $values['dialogTitle'])
            ->set('dialogMessage', $values['dialogMessage'])
            ->set('stayLoggedInButtonText', $values['stayLoggedInButtonText'])
            ->set('logoutNowButtonText', $values['logoutNowButtonText'])
            ->set('debugJavascript', $values['debugJavascript'])
            ->save();

        parent::submitForm($form, $form_state);

    }

}
