<?php

declare(strict_types=1);

/**
 * @file
 * Contains \Drupal\logout_timeout\Service\LogoutTimeoutService
 **/

namespace Drupal\logout_timeout\Service;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\Session\SessionManager;
use Drupal\Core\Url;

/**
 * Defines a LogoutTimeoutManager service.
 */
class LogoutTimeoutService
{

    /**
     * The module handler service.
     *
     * @var \Drupal\Core\Extension\ModuleHandlerInterface
     */
    protected ModuleHandlerInterface $moduleHandler;

    /**
     * The configuration for logout_timeout
     * 
     * @var \Drupal\Core\Config\Config
     */
    protected Config $logoutTimeoutSettings;

    /**
     * The configuration factory service.
     * 
     * @var \Drupal\Core\Config\ConfigFactoryInterface
     */
    protected ConfigFactoryInterface $configFactory;

    /**
     * The session.
     *
     * @var \Drupal\Core\Session\SessionManager
     */
    protected SessionManager $session;

    /**
     * The current user.
     *
     * @var \Drupal\Core\Session\AccountInterface
     */
    protected AccountInterface $currentUser;

    /**
     * Constructs the RiseLogoutService
     * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
     *     The module handler.
     * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
     *     The config factory.
     * @param \Drupal\Core\Session\SessionManager $sessionManager
     *     The session.
     * @param \Drupal\Core\Session\AccountInterface $current_user
     *     Data of the current user
     */

    public function __construct(ModuleHandlerInterface $module_handler, ConfigFactoryInterface $configFactory, SessionManager $sessionManager, AccountInterface $current_user)
    {
        $this->moduleHandler         = $module_handler;
        $this->logoutTimeoutSettings = $configFactory->get('logout_timeout.settings');
        $this->session               = $sessionManager;
        $this->currentUser           = $current_user;
    }

    /**
     * Get the URL of the configured logout target
     * 
     * @return string
     */
    public function getTargetLogoutUrl() : string
    {
        return $this->logoutTimeoutSettings->get('targetLogoutUrl');
    }

    /**
     * Get the configuration value for checking SimpleSAMLphp Auth
     * 
     * @return int
     */
    public function getCheckSamlAuth() : int
    {
        return $this->logoutTimeoutSettings->get('checkSamlAuth');
    }

    /**
     * Check the installation status of the hook_event_dispatcher sub-module user_event_dispatcher.
     * 
     * @return boolean
     */
    public function getHookUserDispatchStatus() : bool
    {
        return $this->moduleHandler->moduleExists('user_event_dispatcher') ? true : false;
    }

    /**
     * Check the installation status of the simplesamlphp_auth module.
     * 
     * @return boolean
     */
    public function getSimpleSamlPhpAuthStatus() : bool
    {
        return $this->moduleHandler->moduleExists('simplesamlphp_auth') ? true : false;
    }

    /**
     * Standard logout for local accounts.
     */
    public function logout() : void
    {

        $user = $this->currentUser;

        $userIsAuthenticated = $user->isAuthenticated();
        $userIsAnonymous     = $user->isAnonymous();

        if (!$userIsAnonymous && $userIsAuthenticated) {
            user_logout();
        }

    }

    /**
     * Logout for accounts authenticated using SAML Auth.
     * 
     * @param int    $showMessage
     * @param string $redirect_url
     */
    public function samlLogout(int $showMessage, string $redirect_url) : void
    {

        $user = $this->currentUser;
        $url  = Url::fromUserInput(
            $redirect_url,
            [
                'absolute' => TRUE,
                'query' => [
                    'logout_timeout' => $showMessage,
                ],
            ]
        );

        \Drupal::logger('logout_timeout')->info('Session closed for %name.', ['%name' => $user->getAccountName()]);

        $simplesaml = \Drupal::service('simplesamlphp_auth.manager');
        $session    = \Drupal::service('session_manager');
        /**
         * Logout procedure was copied from simplesamlphp_auth_user_logout().
         * SimpleSAML Auth originally copied this from the user.module user_logout().
         * An event dispatcher has been added into this procedure, and checks to make sure
         * the user_event_dispatcher module has been installed.
         */
        $session->destroy();
        if (
            $this->getHookUserDispatchStatus() 
            && $this->logoutTimeoutSettings->get('checkHookEventsDispatch')
        ) {
            \Drupal::logger('logout_timeout')->info('Firing the user event for logout of %name.', ['%name' => $user->getAccountName()]);
            $event            = new \Drupal\user_event_dispatcher\Event\User\UserLogoutEvent($user);
            $event_dispatcher = \Drupal::service('event_dispatcher');
            $event_dispatcher->dispatch($event, 'hook_event_dispatcher.user.logout');
        }
        $user->setAccount(new AnonymousUserSession());
        $simplesaml->logout($url->toString());

    }

}
